/* $Workfile: otxdrv_cs.c $   $Revision: 5 $    $Modtime: 5/24/01 12:10p $  */
/*======================================================================

    OTX PCMCIA client driver

    The contents of this file are subject to the Mozilla Public
    License Version 1.1 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a copy of
    the License at http://www.mozilla.org/MPL/

    Software distributed under the License is distributed on an "AS
    IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
    implied. See the License for the specific language governing
    rights and limitations under the License.

    The initial developer of the original code is David A. Hinds
    <dahinds@users.sourceforge.net>.  Portions created by David A. Hinds
    are Copyright (C) 1999 David A. Hinds.  All Rights Reserved.

    Alternatively, the contents of this file may be used under the
    terms of the GNU Public License version 2 (the "GPL"), in which
    case the provisions of the GPL are applicable instead of the
    above.  If you wish to allow the use of your version of this file
    only under the terms of the GPL and not to allow others to use
    your version of this file under the MPL, indicate your decision
    by deleting the provisions above and replace them with the notice
    and other provisions required by the GPL.  If you do not delete
    the provisions above, a recipient may use your version of this
    file under either the MPL or the GPL.

    This code is based on the dummy_cs.c driver written by David
    A. Hinds. It was modified by Arun Dharankar to support Odin 
    TeleSystem's OTX PCMCIA boards. Further modifications have been 
    made by Odin TeleSystems.
    
======================================================================*/

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/system.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ds.h>
#include <pcmcia/bus_ops.h>

#include "otxdrv_cs.h"

/*
   All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
   you do not define PCMCIA_DEBUG at all, all the debug code will be
   left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
   be present but disabled -- but it can then be enabled for specific
   modules at load time with a 'pc_debug=#' option to insmod.
*/

//#define PCMCIA_DEBUG 5

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
static char *version =
//"odints_cs.c 1.29 2001/03/08 14:29:33 (Arun Dharankar)";
// " OTX_DRV_CS_VERSION "
"otxdrv_cs.c (Arun Dharankar/Hanz Johansson)";
#else
#define DEBUG(n, args...)
#endif


/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* Release IO ports after configuration? */
static int free_ports = 0;

/* The old way: bit map of interrupts to choose from */
/* This means pick from 15, 14, 12, 11, 10, 9, 7, 5, 4, and 3 */
static u_int irq_mask = 0xdeb8;
/* Newer, simpler way of listing specific interrupts */
static int irq_list[4] = { -1 };

MODULE_PARM(free_ports, "i");
MODULE_PARM(irq_mask, "i");
MODULE_PARM(irq_list, "1-4i");

/*====================================================================*/

// Interface function between the this PCMCIA Client driver and the OTX Hardware Driver
int OsaLinuxCsIoControl(
  IN void *pDevice, 
  IN unsigned int nCode, 
  IN void *sIn, 
  IN unsigned int cbSin,
  IN OUT void *sOut, 
  IN unsigned int cbSout
);


/*
   The event() function is this driver's Card Services event handler.
   It will be called by Card Services when an appropriate card status
   event is received.  The odints_config() and odints_release() entry points are
   used to configure or release a socket, in response to card
   insertion and ejection events.  They are invoked from the odints
   event handler. 
*/

static void odints_config(dev_link_t *link);
static void odints_release(u_long arg);
static int odints_event(event_t event, int priority,
		       event_callback_args_t *args);

static int odints_cs_set_irq(
  OsaPcmciaBoardNotificationFnP pfnBoardNotificationCallback
);


/*
   The odints_attach() and odints_detach() entry points are used to create and destroy
   "instances" of the driver, where each instance represents everything
   needed to manage one actual PCMCIA card.
*/

static dev_link_t *odints_attach(void);
static void odints_detach(dev_link_t *);

static void odints_interrupt(int , void *, struct pt_regs *);

/*
   You'll also need to prototype all the functions that will actually
   be used to talk to your device.  See 'memory_cs' for a good example
   of a fully self-sufficient driver; the other drivers rely more or
   less on other parts of the kernel.
*/

/*
   The dev_info variable is the "key" that is used to match up this
   device driver with appropriate cards, through the card configuration
   database.
*/

static dev_info_t dev_info = "otxdrv_cs";

/*
   A linked list of "instances" of the odints device.  Each actual
   PCMCIA card corresponds to one device instance, and is described
   by one dev_link_t structure (defined in ds.h).

   You may not want to use a linked list for this -- for example, the
   memory card driver uses an array of dev_link_t pointers, where minor
   device numbers are used to derive the corresponding array index.
*/

static dev_link_t *dev_list = NULL;


// Keep track of how many OTX cards that are currently inserted. 
// This is needed because we only clear the interrupt callback when
// the last card is removed (the interrupt callback is shared by
// all OTX boards).
static int g_coCardsInserted = 0;

/*
   A dev_link_t structure has fields for most things that are needed
   to keep track of a socket, but there will usually be some device
   specific information that also needs to be kept track of.  The
   'priv' pointer in a dev_link_t structure can be used to point to
   a device-specific private data structure, like this.

   To simplify the data structure handling, we actually include the
   dev_link_t structure in the device's private data structure.

   A driver needs to provide a dev_node_t structure for each device
   on a card.  In some cases, there is only one device per card (for
   example, ethernet cards, modems).  In other cases, there may be
   many actual or logical devices (SCSI adapters, memory cards with
   multiple partitions).  The dev_node_t structures need to be kept
   in a linked list starting at the 'dev' field of a dev_link_t
   structure.  We allocate them in the card's private data structure,
   because they generally shouldn't be allocated dynamically.

   In this case, we also provide a flag to indicate if a device is
   "stopped" due to a power management event, or card ejection.  The
   device IO routines can use a flag like this to throttle IO to a
   card that is not ready to accept it.

   The bus_operations pointer is used on platforms for which we need
   to use special socket-specific versions of normal IO primitives
   (inb, outb, readb, writeb, etc) for card IO.
*/
   
typedef struct local_info_t {
    dev_link_t		link;
    dev_node_t		node;
    int			stop;
    void    *pExternalDevice;  // as the device is known in in the OTX Driver
    struct bus_operations *bus;
} local_info_t;


// Interrupt Callback function (in the OTX Driver)
static OsaPcmciaBoardNotificationFnP    g_pfnOsaBoardNotificationCallback = NULL;
static OsaLinuxAddDeviceCallbackFnP     g_pfnOsaAddDeviceCallback = NULL;
static OsaLinuxRemoveDeviceCallbackFnP  g_pfnOsaRemoveDeviceCallback = NULL;


/*====================================================================*/



/*======================================================================
  @func

======================================================================*/
static void cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err = { func, ret };
    CardServices(ReportError, handle, &err);
}


/*======================================================================
  @func

    odints_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

    The dev_link structure is initialized, but we don't actually
    configure the card at this point -- we wait until we receive a
    card insertion event.
    
======================================================================*/
static dev_link_t *odints_attach(void)
{
  local_info_t *local;
  dev_link_t *link;
  client_reg_t client_reg;
  int ret, i;
  
  DEBUG(0, "odints_attach()\n");
  
  /* Allocate space for private device-specific data */
  local = kmalloc(sizeof(local_info_t), GFP_KERNEL);
  if (!local) return NULL;
  memset(local, 0, sizeof(local_info_t));

  link = &local->link; link->priv = local;
  
  /* Initialize the dev_link_t structure */
  link->release.function = &odints_release;
  link->release.data = (u_long)link;
  
  /* Interrupt setup */
  link->irq.Attributes = IRQ_TYPE_EXCLUSIVE | IRQ_HANDLE_PRESENT;
  link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
  if (irq_list[0] == -1) {
    link->irq.IRQInfo2 = irq_mask;
  } else {
    for (i = 0; i < 4; i++) {
      link->irq.IRQInfo2 |= 1 << irq_list[i];
    }
  }
  

  /*
  General socket configuration defaults can go here.  In this
  client, we assume very little, and rely on the CIS for almost
  everything.  In most clients, many details (i.e., number, sizes,
  and attributes of IO windows) are fixed by the nature of the
  device, and can be hard-wired here.
  */
  link->conf.Attributes |= CONF_ENABLE_IRQ;
  link->conf.Vcc = 50;
  link->conf.IntType = INT_MEMORY_AND_IO;
  link->irq.Handler = odints_interrupt;
  link->irq.Instance = local;  // This will be the dev_instance parameter in Handler call
  link->irq.Attributes |= IRQ_TYPE_DYNAMIC_SHARING|IRQ_FIRST_SHARED ;
  
  /* Register with Card Services */
  link->next = dev_list;
  dev_list = link;
  client_reg.dev_info = &dev_info;
  client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
  client_reg.EventMask =
    CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
    CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
    CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
  client_reg.event_handler = &odints_event;
  client_reg.Version = 0x0210;
  client_reg.event_callback_args.client_data = link;

  DEBUG(1, "odints_attach(): Attaching card (link=0x%p, local=0x%p)\n", link, local);
  DEBUG(6, "odints_attach(): OsaLinuxCsIoControl=0x%p\n", OsaLinuxCsIoControl);

  ret = CardServices(RegisterClient, &link->handle, &client_reg);
  if (ret != CS_SUCCESS) {
    cs_error(link->handle, RegisterClient, ret);
    odints_detach(link);
    return NULL;
  }


  // Look for the callback in the kernel symbol file (we should only have to do this for
  // the first card that is inserted, but it should not hurt to do it again (in case
  // the OtxDrv driver was unexpectedly unloaded)).
  g_pfnOsaAddDeviceCallback = (OsaLinuxAddDeviceCallbackFnP) get_module_symbol("OtxDrv", "OsaLinuxAddDeviceCallback");
  
  if (g_pfnOsaAddDeviceCallback) {
    DEBUG(1, "odints_attach(): Calling (*g_pfnOsaAddDeviceCallback)()\n");
    if ((*g_pfnOsaAddDeviceCallback)((void *)link) != 0) {
      DEBUG(1, "odints_attach(): Error, (*g_pfnOsaAddDeviceCallback)() call failed\n");
      odints_detach(link);
      return NULL;
    }
    
  } else {
    DEBUG(1, "odints_attach(): Error, g_pfnOsaAddDeviceCallback==NULL\n");
    odints_detach(link);
    return NULL;
  }

  // Keep track of how many OTX cards that are currently inserted. 
  g_coCardsInserted++;
  
  return link;
} /* odints_attach */


/*======================================================================
  @func

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/
static void odints_detach(dev_link_t *link)
{
  dev_link_t **linkp;
  local_info_t *local = link->priv;
  
  DEBUG(0, "odints_detach(): Detaching card (link=0x%p)\n", link);
  
  /* Locate device structure */
  for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next) {
    if (*linkp == link) break;
  }

  if (*linkp == NULL) {
    return;
  }
  
  /*
  If the device is currently configured and active, we won't
  actually delete it yet.  Instead, it is marked so that when
  the release() function is called, that will trigger a proper
  detach().
  */
  if (link->state & DEV_CONFIG) {
#ifdef PCMCIA_DEBUG
    printk(KERN_DEBUG "otxdrv_cs: detach postponed, '%s' "
      "still locked\n", link->dev->dev_name);
#endif
    link->state |= DEV_STALE_LINK;
    return;
  }


  // Keep track of how many OTX cards that are currently inserted. 
  g_coCardsInserted--;

  g_pfnOsaRemoveDeviceCallback = (OsaLinuxRemoveDeviceCallbackFnP) get_module_symbol("OtxDrv", "OsaLinuxRemoveDeviceCallback");

  if (g_pfnOsaRemoveDeviceCallback) {
    
    if (local->pExternalDevice == NULL) { // Defensive, just checking....
      DEBUG(1, "odints_detach(): Error, local->pExternalDevice == NULL\n");
    }
    
    DEBUG(1, "odints_detach(): Calling (*g_pfnOsaRemoveDeviceCallback)()\n");
    if ((*g_pfnOsaRemoveDeviceCallback)((void *)local->pExternalDevice) != 1) {
      DEBUG(1, "odints_detach(): Error, (*g_pfnOsaReleaseDeviceCallback)() call failed\n");
    }

    local->pExternalDevice = NULL; // Maybe redundant
    
  } else {
    DEBUG(1, "odints_detach(): Error, g_pfnOsaRemoveDeviceCallback==NULL\n");
  }

  // Keep track of how many cards that is using this module
  MOD_DEC_USE_COUNT;

  
  /* Break the link with Card Services */
  if (link->handle) {
    CardServices(DeregisterClient, link->handle);
  }
  
  /* Unlink device structure, and free it */
  *linkp = link->next;
  /* This points to the parent local_info_t struct */
  kfree(link->priv);
    
} /* odints_detach */



#define CS_CHECK(fn, args...) while ((last_ret=CardServices(last_fn=(fn),args))!=0) goto cs_failed

#define CFG_CHECK(fn, args...) if (CardServices(fn, args) != 0) goto next_entry

  
/*======================================================================
  @func

    odints_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    device available to the system.
    
======================================================================*/
static void odints_config(dev_link_t *link)
{
  client_handle_t handle = link->handle;
  local_info_t *dev = link->priv;
  tuple_t tuple;
  cisparse_t parse;
  int last_fn, last_ret;
  u_char buf[64];
  config_info_t conf;
  win_req_t req;
  memreq_t map;
  cistpl_cftable_entry_t dflt = { 0 };
  
  DEBUG(0, "odints_config(): (link=0x%p)\n", link);
  
  /*
  This reads the card's CONFIG tuple to find its configuration
  registers.
  */
  tuple.DesiredTuple = CISTPL_CONFIG;
  tuple.Attributes = 0;
  tuple.TupleData = buf;
  tuple.TupleDataMax = sizeof(buf);
  tuple.TupleOffset = 0;
  CS_CHECK(GetFirstTuple, handle, &tuple);
  CS_CHECK(GetTupleData, handle, &tuple);
  CS_CHECK(ParseTuple, handle, &tuple, &parse);
  link->conf.ConfigBase = parse.config.base;
  link->conf.Present = parse.config.rmask[0];
  
  /* Configure card */
  link->state |= DEV_CONFIG;
  
  /* Look up the current Vcc */
  CS_CHECK(GetConfigurationInfo, handle, &conf);
  link->conf.Vcc = conf.Vcc;
  
  /*
  In this loop, we scan the CIS for configuration table entries,
  each of which describes a valid card configuration, including
  voltage, IO window, memory window, and interrupt settings.
  
    We make no assumptions about the card to be configured: we use
    just the information available in the CIS.  In an ideal world,
    this would work for any PCMCIA card, but it requires a complete
    and accurate CIS.  In practice, a driver usually "knows" most of
    these things without consulting the CIS, and most client drivers
    will only use the CIS to fill in implementation-defined details.
  */
  tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
  CS_CHECK(GetFirstTuple, handle, &tuple);
  while (1) {
    cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);
    CFG_CHECK(GetTupleData, handle, &tuple);
    CFG_CHECK(ParseTuple, handle, &tuple, &parse);
    
    if (cfg->flags & CISTPL_CFTABLE_DEFAULT) dflt = *cfg;
    if (cfg->index == 0) goto next_entry;
    link->conf.ConfigIndex = cfg->index;
    
    /* Does this card need audio output? */
    if (cfg->flags & CISTPL_CFTABLE_AUDIO) {
      link->conf.Attributes |= CONF_ENABLE_SPKR;
      link->conf.Status = CCSR_AUDIO_ENA;
    }
    
    /* Use power settings for Vcc and Vpp if present */
    /*  Note that the CIS values need to be rescaled */
    if (cfg->vcc.present & (1<<CISTPL_POWER_VNOM)) {
      if (conf.Vcc != cfg->vcc.param[CISTPL_POWER_VNOM]/10000)
        goto next_entry;
    } else if (dflt.vcc.present & (1<<CISTPL_POWER_VNOM)) {
      if (conf.Vcc != dflt.vcc.param[CISTPL_POWER_VNOM]/10000)
        goto next_entry;
    }
    
    if (cfg->vpp1.present & (1<<CISTPL_POWER_VNOM))
      link->conf.Vpp1 = link->conf.Vpp2 =
      cfg->vpp1.param[CISTPL_POWER_VNOM]/10000;
    else if (dflt.vpp1.present & (1<<CISTPL_POWER_VNOM))
      link->conf.Vpp1 = link->conf.Vpp2 =
      dflt.vpp1.param[CISTPL_POWER_VNOM]/10000;
    
    /* Do we need to allocate an interrupt? */
    if (cfg->irq.IRQInfo1 || dflt.irq.IRQInfo1)
      link->conf.Attributes |= CONF_ENABLE_IRQ;
    
    /* IO window settings */
    link->io.NumPorts1 = link->io.NumPorts2 = 0;
    if ((cfg->io.nwin > 0) || (dflt.io.nwin > 0)) {
      cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt.io;
      link->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
      if (!(io->flags & CISTPL_IO_8BIT))
        link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
      if (!(io->flags & CISTPL_IO_16BIT))
        link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
      link->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
      link->io.BasePort1 = io->win[0].base;
      link->io.NumPorts1 = io->win[0].len;
      if (io->nwin > 1) {
        link->io.Attributes2 = link->io.Attributes1;
        link->io.BasePort2 = io->win[1].base;
        link->io.NumPorts2 = io->win[1].len;
      }
      /* This reserves IO space but doesn't actually enable it */
      CFG_CHECK(RequestIO, link->handle, &link->io);
    }
    
    /*
    Now set up a common memory window, if needed.  There is room
    in the dev_link_t structure for one memory window handle,
    but if the base addresses need to be saved, or if multiple
    windows are needed, the info should go in the private data
    structure for this device.
    
      Note that the memory window base is a physical address, and
      needs to be mapped to virtual space with ioremap() before it
      is used.
    */
    if ((cfg->mem.nwin > 0) || (dflt.mem.nwin > 0)) {
      cistpl_mem_t *mem =
        (cfg->mem.nwin) ? &cfg->mem : &dflt.mem;
      req.Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_CM;
      req.Attributes |= WIN_ENABLE;
      req.Base = mem->win[0].host_addr;
      req.Size = mem->win[0].len;
      if (req.Size < 0x1000)
        req.Size = 0x1000;
      req.AccessSpeed = 0;
      link->win = (window_handle_t)link->handle;
      CFG_CHECK(RequestWindow, &link->win, &req);
      map.Page = 0; map.CardOffset = mem->win[0].card_addr;
      CFG_CHECK(MapMemPage, link->win, &map);
    }
    /* If we got this far, we're cool! */
    break;
    
next_entry:
    if (link->io.NumPorts1)
      CardServices(ReleaseIO, link->handle, &link->io);
    CS_CHECK(GetNextTuple, handle, &tuple);
  }
  
  /*
  Allocate an interrupt line.  Note that this does not assign a
  handler to the interrupt, unless the 'Handler' member of the
  irq structure is initialized.
  */
  if (link->conf.Attributes & CONF_ENABLE_IRQ)
    CS_CHECK(RequestIRQ, link->handle, &link->irq);
  
    /*
    This actually configures the PCMCIA socket -- setting up
    the I/O windows and the interrupt mapping, and putting the
    card and host interface into "Memory and IO" mode.
  */
  CS_CHECK(RequestConfiguration, link->handle, &link->conf);
  
  /*
  We can release the IO port allocations here, if some other
  driver for the card is going to loaded, and will expect the
  ports to be available.
  */
  if (free_ports) {
    if (link->io.BasePort1)
      release_region(link->io.BasePort1, link->io.NumPorts1);
    if (link->io.BasePort2)
      release_region(link->io.BasePort2, link->io.NumPorts2);
  }
  
#if 1 // Added by HJ to support more OTX PCMCIA cards than just the Thor-2-PCMCIA card
  {
    u_char buf[255];
    int i;
    char *pszProductNumber;
    
    tuple.DesiredTuple = CISTPL_VERS_1;
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = sizeof(buf);
    tuple.TupleOffset = 0;
    
    sprintf(dev->node.dev_name, "otx-pcmcia-card"); // Defensive, will be overwritten by a string from the VERSION_1 tuple
    
    CS_CHECK(GetFirstTuple, handle, &tuple);
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);
    
    // Parse the Version string for the string containing the product number (HAA), and
    // use that string as the dev_name
    // The OTX driver will later query this driver for the dev_name
    for (i = 0; i < parse.version_1.ns; i++) {
      if ((pszProductNumber = strstr(parse.version_1.str + parse.version_1.ofs[i], "HAA-")) != NULL) {
        dev->node.dev_name[DEV_NAME_LEN-1] = '\0';
        strncpy(dev->node.dev_name, pszProductNumber, DEV_NAME_LEN-1);
        break;
      }
    }
  }
#else
  sprintf(dev->node.dev_name, "odints-thor2pcmcia");
#endif
  /*
  At this point, the dev_node_t structure(s) need to be
  initialized and arranged in a linked list at link->dev.
  */
  dev->node.major = dev->node.minor = 0;
  link->dev = &dev->node;
  
  
  /* Finally, report what we've done */
  printk(KERN_INFO "%s: index 0x%02x: Vcc %d.%d",
    dev->node.dev_name, link->conf.ConfigIndex,
    link->conf.Vcc/10, link->conf.Vcc%10);
  if (link->conf.Vpp1) {
    printk(", Vpp %d.%d", link->conf.Vpp1/10, link->conf.Vpp1%10);
  }
  if (link->conf.Attributes & CONF_ENABLE_IRQ) {
    printk(", irq %d", link->irq.AssignedIRQ);
  }
  if (link->io.NumPorts1) {
    printk(", io 0x%04x-0x%04x", link->io.BasePort1,
      link->io.BasePort1+link->io.NumPorts1-1);
  }
  if (link->io.NumPorts2) {
    printk(" & 0x%04x-0x%04x", link->io.BasePort2,
      link->io.BasePort2+link->io.NumPorts2-1);
  }
  if (link->win) {
    printk(", mem 0x%06lx-0x%06lx", req.Base,
      req.Base+req.Size-1);
  }
  printk("\n");
  
  
  if(!(link->io.NumPorts1 > 0 && link->io.NumPorts1 < 64)) {
    DEBUG(0, "WARNING: size1 (%d) is out of range!\n", link->io.NumPorts1);
  }
  
  if(!(link->io.NumPorts2 > 0 && link->io.NumPorts2 < 64)) {
    DEBUG(0, "WARNING: size2 (%d) is out of range!\n", link->io.NumPorts2);
  }
  
  if(!(link->io.BasePort1 > 0 && link->io.BasePort1 < 0xffff)) {
    DEBUG(0, "WARNING: iobase1 (0x%x) is out of range!\n", link->io.BasePort1);
  }
  
  if(!(link->io.BasePort2 > 0 && link->io.BasePort2 < 0xffff)) {
    DEBUG(0, "WARNING: iobase2 (0x%x) is out of range!\n", link->io.BasePort2);
  }
  
  // Keep track of how many cards that is using this module
  MOD_INC_USE_COUNT;
  
  link->state &= ~DEV_CONFIG_PENDING;
  return;
  
cs_failed:
  cs_error(link->handle, last_fn, last_ret);
  odints_release((u_long)link);

} /* odints_config */


/*======================================================================
  @func

    After a card is removed, odints_release() will unregister the
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
======================================================================*/
static void odints_release(u_long arg)
{
  dev_link_t *link = (dev_link_t *)arg;
  
  DEBUG(0, "odints_release(): (link=0x%p)\n", link);
  
  /*
  If the device is currently in use, we won't release until it
  is actually closed, because until then, we can't be sure that
  no one will try to access the device or its data structures.
  */
  if (link->open) {
    DEBUG(1, "otxdrv_cs: release postponed, '%s' still open\n",
      link->dev->dev_name);
    link->state |= DEV_STALE_CONFIG;
    return;
  }
  
  /* Unlink the device chain */
  link->dev = NULL;
  
  /*
  In a normal driver, additional code may be needed to release
  other kernel data structures associated with this device. 
  */
  
  /* Don't bother checking to see if these succeed or not */
  if (link->win) {
    CardServices(ReleaseWindow, link->win);
  }
  CardServices(ReleaseConfiguration, link->handle);
  if (link->io.NumPorts1) {
    CardServices(ReleaseIO, link->handle, &link->io);
  }
  if (link->irq.AssignedIRQ) {
    CardServices(ReleaseIRQ, link->handle, &link->irq);
  }
  link->state &= ~DEV_CONFIG;
  
  if (link->state & DEV_STALE_LINK) {
    odints_detach(link);
  }

} /* odints_release */

/*======================================================================
  @func

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.

    When a CARD_REMOVAL event is received, we immediately set a
    private flag to block future accesses to this device.  All the
    functions that actually access the device should check this flag
    to make sure the card is still present.
    
======================================================================*/
static int odints_event(
  event_t event, 
  int priority,
	event_callback_args_t *args
)
{
  dev_link_t *link = args->client_data;
  local_info_t *local = link->priv;
  
  DEBUG(1, "odints_event(): card link=0x%p, (event=0x%06x)\n", link, event);
  
  switch (event) {
  case CS_EVENT_CARD_REMOVAL:
    DEBUG(1, "odints_event(): CS_EVENT_CARD_REMOVAL\n");
    link->state &= ~DEV_PRESENT;
    if (link->state & DEV_CONFIG) {
      ((local_info_t *)link->priv)->stop = 1;
      mod_timer(&link->release, jiffies + HZ/20);
    }
    break;
  case CS_EVENT_CARD_INSERTION:
    DEBUG(1, "odints_event(): CS_EVENT_CARD_INSERTION\n");
    link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
    local->bus = args->bus;
    odints_config(link);
    break;
  case CS_EVENT_PM_SUSPEND:
    DEBUG(1, "odints_event(): CS_EVENT_PM_SUSPEND\n");
    link->state |= DEV_SUSPEND;
    /* Fall through... */
  case CS_EVENT_RESET_PHYSICAL:
    /* Mark the device as stopped, to block IO until later */
    DEBUG(1, "odints_event(): CS_EVENT_RESET_PHYSICAL\n");
    local->stop = 1;
    if (link->state & DEV_CONFIG)
      CardServices(ReleaseConfiguration, link->handle);
    break;
  case CS_EVENT_PM_RESUME:
    DEBUG(1, "odints_event(): CS_EVENT_PM_RESUME\n");
    link->state &= ~DEV_SUSPEND;
    /* Fall through... */
  case CS_EVENT_CARD_RESET:
    DEBUG(1, "odints_event(): CS_EVENT_CARD_RESET\n");
    if (link->state & DEV_CONFIG)
      CardServices(RequestConfiguration, link->handle, &link->conf);
    local->stop = 0;
    /*
    In a normal driver, additional code may go here to restore
    the device state and restart IO. 
    */
    break;
  }
  return 0;
} /* odints_event */




/*=============================================================================
  @func

  Interface function between the otxdrv_cs module and the OtxDrv module.

=============================================================================*/
int OsaLinuxCsIoControl(
  IN void *pDevice,       // Target device for this command (pointer to an element in dev_list)
  IN unsigned int nCode, 
  IN void *sIn, 
  IN unsigned int cbSin,
  IN OUT void *sOut, 
  IN unsigned int cbSout
)
{

  switch (nCode) {
  case OTX_DRV_CS_IO_GET_VERSION:
    {
      OsaLinuxCsIoOutGetVersionS *psOut = (OsaLinuxCsIoOutGetVersionS *) sOut;

      if (cbSout < sizeof(OsaLinuxCsIoOutGetVersionS)) {
      	DEBUG(1, "OsaLinuxCsIoControlCall: Error, output buffer too small (OTX_DRV_CS_IO_GET_VERSION)\n");
        return 1; // Error
      }

      psOut->nVersion = OTX_DRV_CS_VERSION;

    }
    break;

  // The OTX driver will ask this driver for the dev_name information given
  // to this card. We return this information in this call. 
  case OTX_DRV_CS_IO_GET_DEV_NAME:
    {
      OsaLinuxCsIoOutGetDevNameS *psOut = (OsaLinuxCsIoOutGetDevNameS *) sOut;

      if (cbSout < sizeof(OsaLinuxCsIoOutGetDevNameS)) {
      	DEBUG(1, "OsaLinuxCsIoControlCall: Error, output buffer too small (OTX_DRV_CS_IO_GET_DEV_NAME)\n");
        return 1; // Error
      }

      if (dev_list == NULL) {
		    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_GET_DEV_NAME): Error, no dev_list\n");
		    return 1; // Error
      }

      if (pDevice == NULL) {
		    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_GET_DEV_NAME): Error, non-NULL pDevice must be supplied\n");
		    return 1; // Error
      } else {
        // Traverse the dev_list and search for pDevice (the search is a defensive way, but safer than using pDevice directly)
        dev_link_t *link;

        for (link = dev_list; link && (link != (dev_link_t *)pDevice); link = link->next);
        if (link == NULL) {
    	    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_GET_DEV_NAME): Error, Could not find pDevice=0x%p in dev_list\n", pDevice);
		      return 1; 
        }

        strncpy(psOut->szDevName, link->dev->dev_name, OTX_DRV_CS_DEV_NAME_LEN-1);
      }

    }
    break;

  // The OTX driver will ask this driver for the IO base address information given
  // to this card. We return this information in this call. 
  case OTX_DRV_CS_IO_GET_IO_BASE:
    {
      OsaLinuxCsIoOutGetIoBaseS *psOut = (OsaLinuxCsIoOutGetIoBaseS *) sOut;

      if (cbSout < sizeof(OsaLinuxCsIoOutGetIoBaseS)) {
      	DEBUG(1, "OsaLinuxCsIoControlCall: Error, output buffer too small (OTX_DRV_CS_IO_GET_IO_BASE)\n");
        return 1; // Error
      }

      if (dev_list == NULL) {
		    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_GET_IO_BASE): Error, no dev_list\n");
		    return 1; // Error
      }

      if (pDevice == NULL) {
		    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_GET_IO_BASE): Error, non-NULL pDevice must be supplied\n");
		    return 1; // Error

      } else {
        // Traverse the dev_list and search for pDevice (the search is a defensive way, but safer than using pDevice directly)
        dev_link_t *link;

        for (link = dev_list; link && (link != (dev_link_t *)pDevice); link = link->next);
        if (link == NULL) {
    	    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_GET_IO_BASE): Error, Could not find pDevice=0x%p in dev_list\n", pDevice);
		      return 1; 
        }

        psOut->nSlot     = 1;  // Not used yet
        psOut->nIobase1  = link->io.BasePort1;
        psOut->nSize1    = link->io.NumPorts1;
        psOut->nIobase2  = link->io.BasePort2;
        psOut->nSize2    = link->io.NumPorts2;
      }

    }
    break;


  // The OTX driver will ask this driver for device pointer for the first inserted card.
  // We return this information in this call. 
  case OTX_DRV_CS_IO_GET_CS_DEVICE:
    {
      OsaLinuxCsIoOutGetCsDeviceS *psOut = (OsaLinuxCsIoOutGetCsDeviceS *) sOut;

      if (cbSout < sizeof(OsaLinuxCsIoOutGetCsDeviceS)) {
      	DEBUG(1, "OsaLinuxCsIoControlCall: Error, output buffer too small (OTX_DRV_CS_IO_GET_CS_DEVICE)\n");
        return 1; // Error
      }

      if (dev_list == NULL) {
		    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_GET_CS_DEVICE): Error, no dev_list\n");
		    return 1; // Error
      }

      if (pDevice == NULL) {
        // No specific device given => Get the pointer to the first device in dev_list
        psOut->pCsDevice = (void *) dev_list;
      } else {
		    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_GET_CS_DEVICE): Not supported\n");
		    return 1; // Error, not supported
      }

    }
    break;


  // The OTX driver will provide this driver with a pointer to his implementation of this device
  // (pExternalDevice)
  case OTX_DRV_CS_IO_SET_EXTERNAL_DEVICE:
    {
      OsaLinuxCsIoInSetExternalDeviceS *psIn = (OsaLinuxCsIoInSetExternalDeviceS *) sIn;

      if (dev_list == NULL) {
		    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_SET_EXTERNAL_DEVICE): Error, no dev_list\n");
		    return 1; // Error
      }

      if (pDevice == NULL) {
        // No specific device given 
  	    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_SET_EXTERNAL_DEVICE): pDevice must point to a valid device\n");
		    return 1; // Error, not supported
      } else {
        // Traverse the dev_list and search for pDevice (the search is a defensive way, but safer than using pDevice directly)
        dev_link_t *link;
        local_info_t *local;

        for (link = dev_list; link && (link != (dev_link_t *)pDevice); link = link->next);
        if (link == NULL) {
    	    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_GET_IO_BASE): Error, Could not find pDevice=0x%p in dev_list\n", pDevice);
		      return 1; 
        }

        local = link->priv;

  	    DEBUG(0, "OsaLinuxCsIoControlCall(OTX_DRV_CS_IO_SET_EXTERNAL_DEVICE): link=0x%p, setting pExternalDevice=0x%p\n", link, psIn->pExternalDevice);

        local->pExternalDevice = psIn->pExternalDevice;
#if 0 // Temporary, should be removed
      	g_pExternalDevice = local->pExternalDevice;
#endif
	    }


    }
    break;

  case OTX_DRV_CS_IO_SET_INTERRUPT_CALLBACK:
    {
      OsaLinuxCsIoInSetInterruptCallbackS *psIn = (OsaLinuxCsIoInSetInterruptCallbackS *) sIn;
      int rc;

    	// pDevice is not used in this call (since all devices share the same 
      // BoardNotificationCallback function

      rc = odints_cs_set_irq(psIn->pfnBoardNotificationCallback);

      if (rc) {
      	DEBUG(1, "OsaLinuxCsIoControlCall: Error, OTX_DRV_CS_IO_SET_INTERRUPT_CALLBACK call failed\n");
        return 1; // Error
      }
    }
    break;

  case OTX_DRV_CS_IO_CLR_CALLBACKS:
    {

    	// pDevice is not used in this call (since all devices share the same 
      // BoardNotificationCallback function

      if (g_coCardsInserted > 0) {
        // Do nothing. Only clear the Interrupt Callback function when the last card is removed
        DEBUG(1, "OsaLinuxCsIoControlCall(): Non-Last card => Skipping clearing of callback functions\n");
        return 2;  // Not an error
      } else {
        // Last card. We need to clear the callback functions
        // When the OTX driver unloads we can no longer make calls to him (e.g. we should no 
        // longer call his interrupt callback function. 
        // Therefore we null all the callback function pointers here. 

      	g_pfnOsaBoardNotificationCallback = NULL;
        g_pfnOsaAddDeviceCallback = NULL;
        g_pfnOsaRemoveDeviceCallback = NULL;

      }

    }
    break;

  }

	DEBUG(1, "OsaLinuxCsIoControlCall: nCode=0x%x returns ok\n", (int)nCode);
	return 0;

}

/*=============================================================================
  @func

   The OTX driver will supply this driver with the interrupt callback function 
   to be called for every interrupt. This function pointer is supplied in this call.

=============================================================================*/
static int odints_cs_set_irq(
  OsaPcmciaBoardNotificationFnP pfnBoardNotificationCallback
) 
{
  if (g_pfnOsaBoardNotificationCallback == NULL) {
	  g_pfnOsaBoardNotificationCallback = pfnBoardNotificationCallback;
	  DEBUG(1, "odints_cs_set_irq: g_pfnOsaBoardNotificationCallback set (0x%x) \n", (int)g_pfnOsaBoardNotificationCallback);
  } else {
	  DEBUG(1, "odints_cs_set_irq: g_pfnOsaBoardNotificationCallback already set. Call ignored.\n");
  }
	return 0;
}

/*=============================================================================
  @func

  Interrupt Service Routine
  
  When called, we pass on the call to the OTX Driver Interrupt callback
  function (if such a function has be supplied to us).

=============================================================================*/
static void odints_interrupt(
  int irq, 
  void *dev_instance, 
  struct pt_regs *regs
)
{
  local_info_t *local = (local_info_t *)dev_instance;

	DEBUG(0, "odints_interrupt: irq=%d local=0x%x regs=0x%x\n",
		irq, (int)dev_instance, (int)regs);

  if (g_pfnOsaBoardNotificationCallback == NULL) {
		DEBUG(0, "odints_interrupt: called with no callback!\n");
		return;
	}

  if (local == NULL) {
		DEBUG(0, "odints_interrupt: Unexpected, called with NULL instance!\n");
		return;
  }

  if (local->pExternalDevice == NULL) {
		DEBUG(0, "odints_interrupt: called with no local->pExternalDevice!\n");
		return;
	}

	DEBUG(6, "odints_interrupt: pExternalDevice=0x%p\n", local->pExternalDevice);

  // Call OTX Driver Interrupt callback function (which will serve the interrupt)
	(*g_pfnOsaBoardNotificationCallback)(irq, local->pExternalDevice, regs);

}






/*=============================================================================
  @func

  Driver Entry function
  
  Called once when the driver loads.

=============================================================================*/
static int __init init_odints_cs(void)
{
  servinfo_t serv;
  DEBUG(0, "init_odints_cs(): %s\n", version);
  CardServices(GetCardServicesInfo, &serv);
  if (serv.Revision != CS_RELEASE_CODE) {
    printk(KERN_NOTICE "otxdrv_cs: Card Services release "
      "does not match!\n");
    return -1;
  }
  register_pccard_driver(&dev_info, &odints_attach, &odints_detach);
  return 0;
}

/*=============================================================================
  @func

  Driver Exit function
  
  Called once when the driver unloads.

=============================================================================*/
static void __exit exit_odints_cs(void)
{
  DEBUG(0, "otxdrv_cs: unloading\n");
  unregister_pccard_driver(&dev_info);
  while (dev_list != NULL) {
    del_timer(&dev_list->release);
    if (dev_list->state & DEV_CONFIG)
      odints_release((u_long)dev_list);
    odints_detach(dev_list);
  }
}

module_init(init_odints_cs);
module_exit(exit_odints_cs);
